//-----------------------------------------------------------------------------
//  (c) 2006 by Basler Vision Technologies
//  Section: Vision Components
//  Project: GenApi
//  Author:  Alexander Happe
//  $Header$
//
//  License: This file is published under the license of the EMVA GenICam  Standard Group.
//  A text file describing the legal terms is included in  your installation as 'GenICam_license.pdf'.
//  If for some reason you are missing  this file please contact the EMVA or visit the website
//  (http://www.genicam.org) for a full copy.
//
//  THIS SOFTWARE IS PROVIDED BY THE EMVA GENICAM STANDARD GROUP "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
//  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
//  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE EMVA GENICAM STANDARD  GROUP
//  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  SPECIAL,
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT  LIMITED TO,
//  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  DATA, OR PROFITS;
//  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  THEORY OF LIABILITY,
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT  (INCLUDING NEGLIGENCE OR OTHERWISE)
//  ARISING IN ANY WAY OUT OF THE USE  OF THIS SOFTWARE, EVEN IF ADVISED OF THE
//  POSSIBILITY OF SUCH DAMAGE.
//-----------------------------------------------------------------------------
/*!
\file
\brief    Definition of the FloatT class template
\ingroup GenApi_Implementation
*/

#ifndef GENAPI_FLOATT_H
#define GENAPI_FLOATT_H

#include "../NodeCallback.h"
#include "INodePrivate.h"
#include "../Synch.h"
#include "../Counter.h"
#include "Value2String.h"
#include "NodeMacros.h"
#include "GenApi/Log.h"
#include <limits>

namespace GenApi
{
    /**
    \brief Implementation of the IFloat Interface
    */
    template<class Base>
    class GENAPI_DECL FloatT : public Base
    {
    public:

        //! Constructor
        FloatT< Base >() :
            Base(),
            m_ImposedMax( (std::numeric_limits<double>::max)() ),
            m_ImposedMin( -(std::numeric_limits<double>::max)() )
        {
        }

        //! Implementation of IFloat::SetValue()
        virtual void SetValue(double Value, bool Verify = true)
        {

            // a list of callbacks to fire held outside(!) the autolock on the stack(!)
            std::list<CNodeCallback*> CallbacksToFire;
            {
                AutoLock l(Base::GetLock());

                GCLOG_INFO_PUSH(Base::m_pValueLog)("SetValue( %f )...", Value );

                if( Verify )
                {
                    if( !IsWritable( this ) )
                        throw ACCESS_EXCEPTION("Node is not writable");

                    CHECK_RANGE_FLT(Value, Base::InternalGetMin(), Base::InternalGetMax());
                }

                {
                    typename Base::PostSetValueFinalizer PostSetValueCaller(this, CallbacksToFire);  // dtor calls Base::PostSetValue

                    // invalidates all nodes if this is the first call in a chain of SetValue calls
                    Base::PreSetValue();

                    // sets the value
                    Base::InternalSetValue(Value, Verify);

                    if( Verify )
                        Base::InternalCheckError();

                    #if ! defined( DISABLE_VALUE_CACHING ) || (DISABLE_VALUE_CACHING == 0)
                        // Fill cache
                        if( WriteThrough == static_cast<INode *>(this)->GetCachingMode() )
                        {
                            m_ValueCache = Value;
                            Base::m_ValueCacheValid = true;
                        }
                    #endif

                }

                GCLOG_INFO_POP(Base::m_pValueLog)("...SetValue" );

                // fire callbacks inside the lock
                std::list<CNodeCallback*>::iterator ptrCallback;
                for( ptrCallback = CallbacksToFire.begin(); ptrCallback != CallbacksToFire.end(); ptrCallback++ )
                {
                    (*ptrCallback)->operator ()(cbPostInsideLock);
                }
            }

            // fire callbacks outside the lock
            std::list<CNodeCallback*>::iterator ptrCallback;
            for( ptrCallback = CallbacksToFire.begin(); ptrCallback != CallbacksToFire.end(); ptrCallback++ )
            {
                (*ptrCallback)->operator ()(cbPostOutsideLock);
            }
        }

        //! Implementation of IFloat::operator=()
        virtual IFloat& operator=(double Value)
        {
            SetValue(Value);

            return *this;
        }

        //! Get value of the node as string
        virtual GenICam::gcstring InternalToString(bool Verify = false, bool IgnoreCache = false)
        {
            AutoLock l(Base::GetLock());

            std::ostringstream Buffer;

            switch( Base::InternalGetDisplayNotation() )
            {
            case fnAutomatic:
#ifdef _MSC_VER
                Buffer.setf(0,std::ios::floatfield);
#else
                Buffer.setf((std::_Ios_Fmtflags)0,std::ios::floatfield);
#endif
                break;
            case fnFixed:
                Buffer.setf(std::ios::fixed, std::ios::floatfield);
                break;
            case fnScientific:
                Buffer.setf(std::ios::scientific, std::ios::floatfield);
                break;
            #pragma BullseyeCoverage off
            default:
                assert(false);
            #pragma BullseyeCoverage on
            }

            if( Base::InternalGetDisplayPrecision() > 0 )
                Buffer << std::setprecision( (std::streamsize)Base::InternalGetDisplayPrecision()) << GetValue( Verify );
            else
                Buffer << GetValue( Verify, IgnoreCache );

            GenICam::gcstring valueString;
            valueString = Buffer.str().c_str();
            return valueString;
        }

        //! Implementation of IFloat::GetValue()
        virtual double GetValue(bool Verify = false, bool IgnoreCache = false)
        {
            AutoLock l(Base::GetLock());

            // Note that readability is tested regardless of Verify
            if( !IsReadable( this ) )
                throw ACCESS_EXCEPTION("Node is not readable");

            #if ! defined( DISABLE_VALUE_CACHING ) || (DISABLE_VALUE_CACHING == 0)
                // Use cache
                if (!IgnoreCache && Base::m_ValueCacheValid && !Verify)
                {
                    GCLOG_INFO(Base::m_pValueLog)("GetValue = %f  (from cache)", m_ValueCache );

                    return m_ValueCache;
                }
            #endif

            GCLOG_INFO_PUSH(Base::m_pValueLog)("GetValue...");

            const double Value( Base::InternalGetValue(Verify, IgnoreCache) );

            GCLOG_INFO_POP(Base::m_pValueLog)("...GetValue = %f", Value );

            if( Verify )
            {
                CHECK_RANGE_FLT(Value, Base::InternalGetMin(), Base::InternalGetMax());
                Base::InternalCheckError();
            }

            #if ! defined( DISABLE_VALUE_CACHING ) || (DISABLE_VALUE_CACHING == 0)
                // Fill cache
                const ECachingMode CachingMode(static_cast<INode *>(this)->GetCachingMode());
                if(    WriteThrough == CachingMode
                    || WriteAround  == CachingMode  )
                {
                    m_ValueCache = Value;
                    Base::m_ValueCacheValid = true;
                }
            #endif

            return Value;

        }

        //! Implementation of IFloat::operator()()
        virtual double operator()()
        {
            return GetValue();
        }

        //! Implementation of IFloat::operator*()
        virtual double operator*()
        {
            return GetValue();
        }

        //! Set value of the node as string
        virtual void InternalFromString(const GenICam::gcstring& ValueStr, bool Verify = true)
        {
            double value;
            if (!String2Value(ValueStr,&value))
                throw INVALID_ARGUMENT_EXCEPTION("Node '%s' : cannot convert string '%s' to double", Base::m_Name.c_str(),  ValueStr.c_str() );

            SetValue(value, Verify);
        }


        //! Implementation of IFloat::GetMin()
        virtual double GetMin()
        {
            AutoLock l(Base::GetLock());

            GCLOG_INFO_PUSH(Base::m_pRangeLog)("GetMin...");

            double Minimum = Base::InternalGetMin();
            Minimum = (std::max)(Minimum, m_ImposedMin);

            GCLOG_INFO_POP(Base::m_pRangeLog)("...GetMin = %f", Minimum );

            return Minimum;

        }

        //! Implementation of IFloat::GetMax()
        virtual double GetMax()
        {
            AutoLock l(Base::GetLock());

            GCLOG_INFO_PUSH(Base::m_pRangeLog)("GetMax...");

            double Maximum(Base::InternalGetMax());
            Maximum = (std::min)( Maximum, m_ImposedMax );

            GCLOG_INFO_POP(Base::m_pRangeLog)("...GetMax = %f", Maximum );

            return Maximum;

        }

        //! True if the float has a constant increment
        virtual bool HasInc()
        {
            AutoLock l(Base::GetLock());

            GCLOG_INFO_PUSH(Base::m_pRangeLog)("HasInc...");

            bool HasInc( Base::InternalHasInc() );

            GCLOG_INFO_POP(Base::m_pRangeLog)("...HasInc = " + HasInc ? "true" : "false" );

            return HasInc;

        }

        //! Get the constant increment if there is any
        virtual double GetInc()
        {
            AutoLock l(Base::GetLock());

            if( !Base::InternalHasInc() )
                throw RUNTIME_EXCEPTION("Node '%s' does not have a constant increment", Base::InternalGetName(false).c_str() );

            GCLOG_INFO_PUSH(Base::m_pRangeLog)("GetInc...");

            double Inc( Base::InternalGetInc() );

            GCLOG_INFO_POP(Base::m_pRangeLog)("...GetInc = %f", Inc );

            return Inc;
        }


        //! Implementation of IFloat::GetRepresentation()
        virtual  ERepresentation GetRepresentation()
        {
            AutoLock l(Base::GetLock());
            return Base::InternalGetRepresentation();
        }

        //! Implementation of IFloat::GetUnit()
        virtual GenICam::gcstring GetUnit() const
        {
            AutoLock l(Base::GetLock());
            return Base::InternalGetUnit();
        }

        //! Get the way the float should be converted to a string
        virtual EDisplayNotation GetDisplayNotation() const
        {
            AutoLock l(Base::GetLock());
            return Base::InternalGetDisplayNotation();
        }

        //! Get the DisplayPrecision to be used when converting the float to a string
        virtual int64_t GetDisplayPrecision() const
        {
            AutoLock l(Base::GetLock());
            return Base::InternalGetDisplayPrecision();
        }

        //! Restrict minimum value
        virtual void ImposeMin(double Value)
        {
            m_ImposedMin = Value;
            Base::SetInvalid(INodePrivate::simAll);
        }

        //! Restrict maximum value
        virtual void ImposeMax(double Value)
        {
            m_ImposedMax = Value;
            Base::SetInvalid(INodePrivate::simAll);
        }

    protected:
        //! The cache for the float value
        double m_ValueCache;

        //! a Value imposed to the maximum
        double m_ImposedMax;

        //! a Value imposed to the minimum
        double m_ImposedMin;

    };
}

#endif // GENAPI_FLOATT_H
