//-----------------------------------------------------------------------------
//  (c) 2006 by Basler Vision Technologies
//  Section: Vision Components
//  Project: GenApi
//  Author:  Fritz Dierks
//  $Header$
//
//  License: This file is published under the license of the EMVA GenICam  Standard Group.
//  A text file describing the legal terms is included in  your installation as 'GenICam_license.pdf'.
//  If for some reason you are missing  this file please contact the EMVA or visit the website
//  (http://www.genicam.org) for a full copy.
//
//  THIS SOFTWARE IS PROVIDED BY THE EMVA GENICAM STANDARD GROUP "AS IS"
//  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
//  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
//  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE EMVA GENICAM STANDARD  GROUP
//  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  SPECIAL,
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT  LIMITED TO,
//  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  DATA, OR PROFITS;
//  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  THEORY OF LIABILITY,
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT  (INCLUDING NEGLIGENCE OR OTHERWISE)
//  ARISING IN ANY WAY OUT OF THE USE  OF THIS SOFTWARE, EVEN IF ADVISED OF THE
//  POSSIBILITY OF SUCH DAMAGE.
//-----------------------------------------------------------------------------
/*!
\file
\brief    Defintion of macros for property conversion map
\ingroup GenApi_Implementation
*/

#ifndef GENAPI_NODEMACROS_H
#define GENAPI_NODEMACROS_H

//*************************************************************
// Macros forming the CONVERSION_MAP
//*************************************************************

typedef enum
{
    In,
    Out
} Direction_t;

//! ---------------------------------------------------------------------
#define BEGIN_CONVERT_MAP \
    virtual bool SetProperty(const GenICam::gcstring& PropertyName, const GenICam::gcstring& ValueStr) \
    { \
        return SetProperty(PropertyName, ValueStr, ""); \
    } \
virtual bool SetProperty(const GenICam::gcstring& PropertyName, const GenICam::gcstring& ValueStr, const GenICam::gcstring& AttributeStr) \
    { \
        bool PropertyAlreadyPresent=false; \
        GenICam::gcstring_vector::iterator ptr; \
        for(ptr=m_PropertyNames.begin(); ptr!=m_PropertyNames.end(); ptr++) \
            if(*ptr == PropertyName) \
                PropertyAlreadyPresent = true; \
        if(!PropertyAlreadyPresent) \
            m_PropertyNames.push_back(PropertyName); \
        GenICam::gcstring _ValueStr=ValueStr; \
        GenICam::gcstring _AttributeStr = AttributeStr; \
        return AccessProperty( PropertyName, _ValueStr, _AttributeStr, In ); \
    } \
virtual bool AccessProperty(const GenICam::gcstring& PropertyName, GenICam::gcstring& ValueStr, GenICam::gcstring& AttributeStr, Direction_t Direction) \
  { \
    if(    Direction == Out ) \
    { \
        ValueStr = ""; \
        AttributeStr = ""; \
    }


//! ---------------------------------------------------------------------
#define CHAIN_CONVERT_MAP(TARGET) \
  if(TARGET::AccessProperty(PropertyName, ValueStr, AttributeStr, Direction)) \
    return true;


//! ---------------------------------------------------------------------
#define SWITCH_CONVERT_MAP \
    int lID = GetIDFromMap( PropertyName ); \
    switch ( lID ) \
    {


//! ---------------------------------------------------------------------
#define CONVERT_ENUM_ENTRY(_ID, Member, _EnumDelegate) \
    case _ID: \
    { \
        if (Direction == In) \
        { \
            if( !_EnumDelegate::FromString(ValueStr, &(Member))) \
                throw PROPERTY_EXCEPTION("%s : property '%s' : cannot convert value '%s'", m_Name.c_str(), PropertyName.c_str(), ValueStr.c_str() ); \
            return true; \
        } \
        else if (Direction == Out) \
        { \
            ValueStr = _EnumDelegate::ToString(Member); \
            return true; \
        } \
    } \
    break;

//! ---------------------------------------------------------------------
#define CONVERT_NODESET_REFERENCE(_ID, Member, Interface, ContainerType ) \
    case _ID: \
    { \
        if (Direction == In) \
        { \
            INode * const pNode = m_pNodeMap->GetNode(ValueStr); \
            if(!pNode) \
                throw PROPERTY_EXCEPTION("%s : property '%s' : node '%s' does not exist", dynamic_cast<INode*>(this)->GetName().c_str(), PropertyName.c_str(), ValueStr.c_str() ); \
            Interface * const pListElem = dynamic_cast<Interface*>(pNode); \
            if(!pListElem) \
                throw PROPERTY_EXCEPTION("%s : property list element '%s' : node '%s' has no interface '" #Interface "'", dynamic_cast<INode*>(this)->GetName().c_str(), PropertyName.c_str(), ValueStr.c_str() ); \
            Member.insert(pListElem);\
            return true; \
        } \
        else if (Direction == Out) \
        { \
            ContainerType::iterator ptr; \
            for(ptr=Member.begin(); ptr!=Member.end(); ptr++ ) \
                ValueStr += (*ptr)->GetName() + "\t"; \
            if(ValueStr.size() > 1) \
                ValueStr = ValueStr.substr(0, ValueStr.size()-1); \
            return true; \
        } \
    } \
    break;

//! ---------------------------------------------------------------------
#define CONVERT_NODE_REFERENCE(_ID, Member, Interface) \
    case _ID: \
    { \
        if (Direction == In) \
        { \
            INode * const pNode = m_pNodeMap->GetNode(ValueStr); \
            if(!pNode) \
                throw PROPERTY_EXCEPTION("%s : property '%s' : node '%s' does not exist", dynamic_cast<INode*>(this)->GetName().c_str(), PropertyName.c_str(), ValueStr.c_str() ); \
            Member = dynamic_cast<Interface*>(pNode); \
            if(!Member) \
                throw PROPERTY_EXCEPTION("%s : property '%s' : node '%s' has no interface '" #Interface "'", dynamic_cast<INode*>(this)->GetName().c_str(), PropertyName.c_str(), ValueStr.c_str() ); \
            if (dynamic_cast<INode*>(this) == pNode) \
            throw PROPERTY_EXCEPTION("%s : property '%s' : self reference ", dynamic_cast<INode*>(this)->GetName().c_str(), PropertyName.c_str() ); \
            AddChild(pNode ); \
            return true; \
        } \
        else if (Direction == Out) \
        { \
            INode *pNode = (INode*)Member; \
            if(pNode) \
                ValueStr = pNode->GetName(); \
            return true; \
        } \
    } \
    break;


//! ---------------------------------------------------------------------
#define CONVERT_STRING_ENTRY(_ID, Member) \
    case _ID: \
    { \
        if (Direction == In) \
        { \
            Member = ValueStr; \
            return true; \
        } \
        else if (Direction == Out) \
        { \
            ValueStr = Member; \
            return true; \
        } \
    } \
    break;


//! ---------------------------------------------------------------------
#define CONVERT_NODELIST_REFERENCE2(_ID, Member, Interface) \
    case _ID: \
    { \
        if (Direction == In) \
        { \
            INode * const pNode = m_pNodeMap->GetNode(ValueStr); \
            if(!pNode) \
                throw PROPERTY_EXCEPTION("%s : property '%s' : node '%s' does not exist", dynamic_cast<INode*>(this)->GetName().c_str(), PropertyName.c_str(), ValueStr.c_str() ); \
            Interface * const pListElem = dynamic_cast<Interface*>(pNode); \
            if(!pListElem) \
                throw PROPERTY_EXCEPTION("%s : property list element '%s' : node '%s' has no interface '" #Interface "'", dynamic_cast<INode*>(this)->GetName().c_str(), PropertyName.c_str(), ValueStr.c_str() ); \
            Member.push_back(pListElem);\
            return true; \
        } \
        else if (Direction == Out) \
        { \
            FeatureList_t::iterator ptr; \
            for(ptr=Member.begin(); ptr!=Member.end(); ptr++ ) \
            { \
                ValueStr += (*ptr)->GetNode()->GetName() + "\t"; \
            } \
            if(ValueStr.size() > 1) \
                ValueStr = ValueStr.substr(0, ValueStr.size()-1); \
            return true; \
        } \
    } \
    break;


//! ---------------------------------------------------------------------
#define CONVERT_INVALIDATOR( _ID ) \
    case _ID: \
    { \
        if (Direction == In) \
        { \
            INode *pNode = m_pNodeMap->GetNode( ValueStr ); \
            if (! pNode) \
                throw PROPERTY_EXCEPTION("%s : property '%s' : node '%s' does not exist", dynamic_cast<INode*>(this)->GetName().c_str(), PropertyName.c_str(), ValueStr.c_str() ); \
            m_Invalidators.push_back( pNode ); \
            m_DependingChildren.insert( dynamic_cast< INodePrivate*>(pNode) ); \
            return true; \
        } \
        else if (Direction == Out) \
        { \
            NodeList_t::iterator ptr; \
            for(ptr=m_Invalidators.begin(); ptr!=m_Invalidators.end(); ptr++ ) \
            { \
                ValueStr += (*ptr)->GetName() + "\t"; \
            } \
            if(ValueStr.size() > 1) \
                ValueStr = ValueStr.substr(0, ValueStr.size()-1); \
            return true; \
        } \
    } \
    break;


//! ---------------------------------------------------------------------
#define CONVERT_ENTRY(_ID, Member) \
    case _ID: \
    { \
        if (Direction == In) \
        { \
            if( !String2Value(ValueStr, &Member)) \
                throw PROPERTY_EXCEPTION("%s : property '%s' : cannot convert value '%s'", m_Name.c_str(), PropertyName.c_str(), ValueStr.c_str() ); \
            return true; \
        } \
        else if (Direction == Out) \
        { \
            Value2String(Member, ValueStr); \
            return true; \
        } \
    } \
    break;


//! ---------------------------------------------------------------------
#define CONVERT_REFERENCE(_ID, Member, Interface) \
    case _ID: \
    { \
        if (Direction == In) \
        { \
            INode * const pNode = m_pNodeMap->GetNode(ValueStr); \
            if(!pNode) \
                throw PROPERTY_EXCEPTION("%s : property '%s' : node '%s' does not exist", dynamic_cast<INode*>(this)->GetName().c_str(), PropertyName.c_str(), ValueStr.c_str() ); \
            Member = dynamic_cast<Interface*>(pNode); \
            if(!Member) \
                throw PROPERTY_EXCEPTION("%s : property '%s' : node '%s' has no interface '" #Interface "'", dynamic_cast<INode*>(this)->GetName().c_str(), PropertyName.c_str(), ValueStr.c_str() ); \
            return true; \
        } \
        else if (Direction == Out) \
        { \
            INode *pNode = (INode*)Member; \
            if(pNode) \
                ValueStr = pNode->GetName(); \
            return true; \
        } \
    } \
    break;


//! ---------------------------------------------------------------------
#define CONVERT_NODELIST_REFERENCE(_ID, Member, Interface, ContainerType ) \
    case _ID: \
    { \
        if (Direction == In) \
        { \
            INode * const pNode = m_pNodeMap->GetNode(ValueStr); \
            if(!pNode) \
                throw PROPERTY_EXCEPTION("%s : property '%s' : node '%s' does not exist", dynamic_cast<INode*>(this)->GetName().c_str(), PropertyName.c_str(), ValueStr.c_str() ); \
            Interface * const pListElem = dynamic_cast<Interface*>(pNode); \
            if(!pListElem) \
                throw PROPERTY_EXCEPTION("%s : property list element '%s' : node '%s' has no interface '" #Interface "'", dynamic_cast<INode*>(this)->GetName().c_str(), PropertyName.c_str(), ValueStr.c_str() ); \
            Member.push_back(pListElem);\
            dynamic_cast<CNodeImpl *>(this)->AddChild(pNode); \
            return true; \
        } \
        else if (Direction == Out) \
        { \
            ContainerType::iterator ptr; \
            for(ptr=Member.begin(); ptr!=Member.end(); ptr++ ) \
                ValueStr += (*ptr)->GetNode()->GetName() + "\t"; \
            if(ValueStr.size() > 1) \
                ValueStr = ValueStr.substr(0, ValueStr.size()-1); \
            return true; \
        } \
    } \
    break;


//! ---------------------------------------------------------------------
#define CONVERT_LIST_ENTRY(_ID, Member, _Type) \
    case _ID: \
    { \
        if (Direction == In) \
        { \
            _Type listElem;\
            if( !String2Value(ValueStr, &listElem)) \
                throw PROPERTY_EXCEPTION("%s : property '%s' : cannot convert value '%s'", m_Name.c_str(), PropertyName.c_str(), ValueStr.c_str() ); \
            Member.push_back(listElem);\
            return true; \
        } \
        else if (Direction == Out) \
       {  Values2String( Member, ValueStr ); return true; } \
        } \
    break;


//! ---------------------------------------------------------------------
#define CONVERT_NODELIST_REFERENCE_AUXREF(_ID, Member, Interface, AuxMember, AuxInterface ) \
    case _ID: \
    { \
        if (Direction == In) \
        { \
            INode * const pNode = m_pNodeMap->GetNode(ValueStr); \
            if(!pNode) \
                throw PROPERTY_EXCEPTION("%s : property '%s' : node '%s' does not exist", dynamic_cast<INode*>(this)->GetName().c_str(), PropertyName.c_str(), ValueStr.c_str() ); \
            Interface * const pListElem = dynamic_cast<Interface*>(pNode); \
            if(!pListElem) \
                throw PROPERTY_EXCEPTION("%s : property list element '%s' : node '%s' has no interface '" #Interface "'", dynamic_cast<INode*>(this)->GetName().c_str(), PropertyName.c_str(), ValueStr.c_str() ); \
            INode * const pAux = m_pNodeMap->GetNode(AttributeStr); \
            if(!pAux) \
                throw PROPERTY_EXCEPTION("%s : property '%s' : attr '%s' does not exist", dynamic_cast<INode*>(this)->GetName().c_str(), PropertyName.c_str(), AttributeStr.c_str() ); \
            AuxInterface * const pListElemAux = dynamic_cast<AuxInterface*>(pAux); \
            if(!pListElemAux) \
                throw PROPERTY_EXCEPTION("%s : property list element '%s' : attr '%s' has no interface '" #AuxInterface "'", dynamic_cast<INode*>(this)->GetName().c_str(), PropertyName.c_str(), AttributeStr.c_str() ); \
            Member.push_back(pListElem);\
            AuxMember.push_back(pListElemAux);\
            dynamic_cast<CNodeImpl*>(this)->AddChild(pNode); \
            dynamic_cast<CNodeImpl*>(this)->AddChild(pAux); \
            return true; \
        } \
        else if (Direction == Out) \
        { \
            IntegerList_t::iterator ptrValue; \
            IntegerList_t::iterator ptrAttribute; \
            for(ptrValue=Member.begin(), ptrAttribute=AuxMember.begin(); \
                ptrValue!=Member.end();  \
                ptrValue++, ptrAttribute++ ) \
            { \
                ValueStr += (*ptrValue)->GetNode()->GetName() + "\t"; \
                AttributeStr += (*ptrAttribute)->GetNode()->GetName() + "\t"; \
            } \
            if(ValueStr.size() > 1) \
            { \
                ValueStr = ValueStr.substr(0, ValueStr.size()-1); \
                AttributeStr = AttributeStr.substr(0, AttributeStr.size()-1); \
            }\
            return true; \
        } \
    } \
    break;


//! ---------------------------------------------------------------------
#define CONVERT_NODELIST_REFERENCE3(_ID, Member, Interface, ContainerType ) \
    case _ID: \
    { \
        if (Direction == In) \
        { \
            INode * const pNode = m_pNodeMap->GetNode(ValueStr); \
            if(!pNode) \
                throw PROPERTY_EXCEPTION("%s : property '%s' : node '%s' does not exist", dynamic_cast<INode*>(this)->GetName().c_str(), PropertyName.c_str(), ValueStr.c_str() ); \
            Interface * const pListElem = dynamic_cast<Interface*>(pNode); \
            if(!pListElem) \
                throw PROPERTY_EXCEPTION("%s : property list element '%s' : node '%s' has no interface '" #Interface "'", dynamic_cast<INode*>(this)->GetName().c_str(), PropertyName.c_str(), ValueStr.c_str() ); \
            Member.push_back(pListElem);\
            return true; \
        } \
        else if (Direction == Out) \
        { \
            ContainerType::iterator ptr; \
            for(ptr=Member.begin(); ptr!=Member.end(); ptr++ ) \
                ValueStr += (*ptr)->GetName() + "\t"; \
            if(ValueStr.size() > 1) \
                ValueStr = ValueStr.substr(0, ValueStr.size()-1); \
            return true; \
        } \
    } \
    break;


//! ---------------------------------------------------------------------
#define END_CONVERT_MAP \
    } \
    return false; \
};


#endif  // GENAPI_NODEMACROS_H
