//-----------------------------------------------------------------------------
//  (c) 2005 by Basler Vision Technologies
//  Section: Vision Components
//  Project: GenApiTest
//    Author:  Hartmut Nebelung
//  $Header$
//-----------------------------------------------------------------------------
/*!
  \file
  Test the implementation of CFloat
*/

#ifdef _MSC_VER
#   include "GenApi/impl/GlobalPragmas.h"  // this must be the first include!
#endif

#include "GenApiTest.h"
#include "CallbackTest.h"
#include "CallbackHelper.h"
#include "TestPort.h"
#include "GenApi/GenApi.h"

#include "GenApi/Test/TestCamera_v1_0.h"
#include "GenApi/Test/TestCamera_v1_1.h"

#include "GenApi/NodeCallbackImpl.h"
#include <bitset>
#include "GenApi/Counter.h"

using namespace std;
using namespace GenICam;
using namespace GenApi;

#undef min

CPPUNIT_TEST_SUITE_REGISTRATION( CallbackTestSuite );

// ---------------------------------------------------------------------------
// the test callback with a counter
// ---------------------------------------------------------------------------
struct CallbackUtility
{
    static void Reset() { m_Count=0; }
    static void Callback( INode* ) { ++m_Count; }
    static uint32_t Count() { return m_Count; }
private:
    static unsigned m_Count;
};
unsigned CallbackUtility::m_Count=0;


// ---------------------------------------------------------------------------
// void CallbackTestSuite::TestCallback01()
// ---------------------------------------------------------------------------
/*!
 * Perform basic tests on a float with static callback function
 */
void CallbackTestSuite::TestCallback01()
{
    // create and initilize node map
    gcstring CameraDescription =
        CDFHeader +

        "<Float Name=\"MyFloat\"> \n"
        "  <Value>0.0</Value> \n"
        "  <Min>0.0</Min> \n"
        "  <Max>100.0</Max> \n"
        "  <Unit>dB</Unit> \n"
        "</Float> \n"

        + CDFFooter;

    CNodeMapRef Camera;
    Camera._LoadXMLFromString(CameraDescription);

    CallbackUtility::Reset();

    CFloatPtr ptrFloat01 = Camera._GetNode( "MyFloat" );
    CPPUNIT_ASSERT( ptrFloat01.IsValid() );

    // register function pointer
    Register( ptrFloat01->GetNode(), &CallbackUtility::Callback );

    // register invalid function pointer - nothing should get broken
    void (*fn)(INode*) = NULL;
    Register( ptrFloat01->GetNode(), fn );

    // register member function
    CallbackObject obj;
    void (CallbackObject::*memfn)(INode*) = NULL;
    Register( ptrFloat01->GetNode(), obj, memfn );

    // check if callbacks are fired
    const uint32_t numCalls(42);
    for (uint32_t i=0; i<numCalls; ++i) ptrFloat01->SetValue( i );
    CPPUNIT_ASSERT_EQUAL( numCalls, CallbackUtility::Count() );

}

// varitant of TestCallback01 for strings
void CallbackTestSuite::TestCallback01String()
{
    if(GenApiSchemaVersion < v1_1)
        return;

    // create and initilize node map
    gcstring CameraDescription =
        CDFHeader +

    "<String Name=\"MyString\">                      \n"
    "    <Value>Alle meine Entchen...</Value>   \n"
    "</String>                                \n"
    "                                         \n"

        + CDFFooter;

    CNodeMapRef Camera;
    Camera._LoadXMLFromString(CameraDescription);

    CallbackUtility::Reset();

    CStringPtr ptrString01 = Camera._GetNode( "MyString" );
    CPPUNIT_ASSERT( ptrString01.IsValid() );

    // register function pointer
    Register( ptrString01->GetNode(), &CallbackUtility::Callback );

    // register invalid function pointer - nothing should get broken
    void (*fn)(INode*) = NULL;
    Register( ptrString01->GetNode(), fn );

    // register member function
    CallbackObject obj;
    void (CallbackObject::*memfn)(INode*) = NULL;
    Register( ptrString01->GetNode(), obj, memfn );

    // check if callbacks are fired
    const uint32_t numCalls(42);
    for (uint32_t i=0; i<numCalls; ++i) ptrString01->SetValue( "...schwimmen auf dem See!" );
    CPPUNIT_ASSERT_EQUAL( numCalls, CallbackUtility::Count() );

}

// varitant of TestCallback01 for register node
void CallbackTestSuite::TestCallback01Register()
{
    // create and initilize node map
    gcstring CameraDescription =
        CDFHeader +

    "                                        \n"
    "<Register Name=\"MyRegister\">            \n"
    "    <Address>0x0000</Address>            \n"
    "    <Length>1</Length>                    \n"
    "    <AccessMode>RW</AccessMode>            \n"
    "    <pPort>Port</pPort>                    \n"
    "</Register>                            \n"
    "                                        \n"
    "<Port Name=\"Port\"/>                    \n"
    "                                         \n"

        + CDFFooter;

    CNodeMapRef Camera;
    Camera._LoadXMLFromString(CameraDescription);

    uint8_t RegVal = 0;
    CTestPort Port;
    Port.CreateEntry(0x0000, RegVal, RW);
    Camera._Connect(&Port, "Port");

    CallbackUtility::Reset();

    CRegisterPtr ptrRegister01 = Camera._GetNode( "MyRegister" );
    CPPUNIT_ASSERT( ptrRegister01.IsValid() );

    // register function pointer
    Register( ptrRegister01->GetNode(), &CallbackUtility::Callback );

    // register invalid function pointer - nothing should get broken
    void (*fn)(INode*) = NULL;
    Register( ptrRegister01->GetNode(), fn );

    // register member function
    CallbackObject obj;
    void (CallbackObject::*memfn)(INode*) = NULL;
    Register( ptrRegister01->GetNode(), obj, memfn );

    // check if callbacks are fired
    const uint32_t numCalls(42);
    for (uint32_t i=0; i<numCalls; ++i) ptrRegister01->FromString ("0xAB");
    CPPUNIT_ASSERT_EQUAL( numCalls, CallbackUtility::Count() );
    RegVal = 17;
    for (uint32_t i=0; i<numCalls; ++i)  ptrRegister01->Set (&RegVal, sizeof(RegVal));
    CPPUNIT_ASSERT_EQUAL( 2*numCalls, CallbackUtility::Count() );

}

// ---------------------------------------------------------------------------
// void CallbackTestSuite::TestCallback02()
// ---------------------------------------------------------------------------
/*!
 * Perform basic tests on Float with two callback objects
 */
void CallbackTestSuite::TestCallback02()
{
    // create and initialize node map
    gcstring CameraDescription =
        CDFHeader +

        "<Float Name=\"MyFloat\"> \n"
        "  <Value>0.0</Value> \n"
        "  <Min>0.0</Min> \n"
        "  <Max>100.0</Max> \n"
        "  <Unit>dB</Unit> \n"
        "</Float> \n"

        + CDFFooter;

    CNodeMapRef Camera;
    Camera._LoadXMLFromString(CameraDescription);

    CFloatPtr ptrFloat01(Camera._GetNode( "MyFloat" ));
    CPPUNIT_ASSERT( ptrFloat01.IsValid() );

    CallbackObject O1, O2;
    CallbackHandleType hCbk1 = Register( ptrFloat01->GetNode(), O1, &CallbackObject::Callback );
    CallbackHandleType hCbk2 = Register( ptrFloat01->GetNode(), O2, &CallbackObject::Callback );

    const uint32_t numCalls = 42;
    for (uint32_t i=0; i<numCalls; ++i) ptrFloat01->SetValue( i );

    CPPUNIT_ASSERT_EQUAL( numCalls, O1.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, O2.Count() );

    // de-register callback
    Deregister(hCbk1);
	CPPUNIT_ASSERT( !ptrFloat01->GetNode()->DeregisterCallback(hCbk1) );

    ptrFloat01->SetValue( 33 );
    CPPUNIT_ASSERT_EQUAL( numCalls, O1.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls+1, O2.Count() );

    // deregister invalid callback handle
    CPPUNIT_ASSERT( 0 == ptrFloat01->GetNode()->DeregisterCallback (CallbackHandleType(9999)) );
}

// ---------------------------------------------------------------------------
// void CallbackTestSuite::TestCallback03()
// ---------------------------------------------------------------------------
/*!
 * Perform basic tests on Float with two callback objects
 */
void CallbackTestSuite::TestCallback03()
{
    // create and initilize node map
    gcstring CameraDescription =
        CDFHeader +

        "<Boolean Name=\"MyBoolean\"> \n"
        "  <Value>true</Value> \n"
        "</Boolean> \n"

        + CDFFooter;

    CNodeMapRef Camera;
    Camera._LoadXMLFromString(CameraDescription);

    CallbackObject O1, O2;

    CBooleanPtr ptrBool(Camera._GetNode( "MyBoolean" ));
    CPPUNIT_ASSERT( ptrBool.IsValid() );

    Register( ptrBool->GetNode(), &CallbackUtility::Callback );
    Register( ptrBool->GetNode(), O1, &CallbackObject::Callback );
    Register( ptrBool->GetNode(), O2, &CallbackObject::Callback );

    const uint32_t numCalls(42);
    CallbackUtility::Reset();

    for (uint32_t i=0; i<numCalls; ++i) ptrBool->SetValue( i&0x1 );

    CPPUNIT_ASSERT_EQUAL( numCalls, CallbackUtility::Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, O1.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, O2.Count() );

}

// ---------------------------------------------------------------------------
// void CallbackTestSuite::TestCallback04()
// ---------------------------------------------------------------------------
/*!
 * Perform basic tests on Command with two callback objects and the callbackutility
 */
void CallbackTestSuite::TestCallback04()
{
    // create and initilize node map
    gcstring CameraDescription =
        CDFHeader +

        "<Command Name=\"MyCommand\"> \n"
        "  <Value>0</Value> \n"
        "  <CommandValue>0xff</CommandValue> \n"
        "</Command> \n"

        + CDFFooter;

    CNodeMapRef Camera;
    Camera._LoadXMLFromString(CameraDescription);

    CallbackObject O1, O2;

    CCommandPtr ptrCmd(Camera._GetNode( "MyCommand" ));
    CPPUNIT_ASSERT( ptrCmd.IsValid() );

    Register( ptrCmd->GetNode(), &CallbackUtility::Callback );
    Register( ptrCmd->GetNode(), O1, &CallbackObject::Callback );
    Register( ptrCmd->GetNode(), O2, &CallbackObject::Callback );

    const uint32_t numCalls(42);
    CallbackUtility::Reset();

    for (uint32_t i=0; i<numCalls; ++i) ptrCmd->Execute();

    CPPUNIT_ASSERT_EQUAL( numCalls, CallbackUtility::Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, O1.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, O2.Count() );

}

// ---------------------------------------------------------------------------
// void CallbackTestSuite::TestCallback05()
// ---------------------------------------------------------------------------
/*!
 * Perform basic tests on Enums with two callback objects and the callbackutility
 */
void CallbackTestSuite::TestCallback05()
{
    // create and initilize node map
    gcstring CameraDescription =
        CDFHeader +

        "<Enumeration Name=\"MyFood\"> \n"
        "  <EnumEntry Name=\"Orange\" > \n"
        "    <Value>0</Value> \n"
        "  </EnumEntry> \n"
        "  <EnumEntry Name=\"Apple\" > \n"
        "    <Value>1</Value> \n"
        "  </EnumEntry> \n"
        "  <EnumEntry Name=\"Pretzel\" > \n"
        "    <Value>2</Value> \n"
        "  </EnumEntry> \n"
        "  <EnumEntry Name=\"Juice\" > \n"
        "    <Value>3</Value> \n"
        "  </EnumEntry> \n"
        "  <Value>0</Value> \n"
        "</Enumeration> \n"

        + CDFFooter;

    CNodeMapRef Camera;
    Camera._LoadXMLFromString(CameraDescription);

    CallbackObject O1, O2;

    CEnumerationPtr ptrEnum(Camera._GetNode( "MyFood" ));
    CPPUNIT_ASSERT( ptrEnum.IsValid() );

    Register( ptrEnum->GetNode(), &CallbackUtility::Callback );
    Register( ptrEnum->GetNode(), O1, &CallbackObject::Callback );
    Register( ptrEnum->GetNode(), O2, &CallbackObject::Callback );

    const uint32_t numCalls(42);
    CallbackUtility::Reset();

    for (uint32_t i=0; i<numCalls; ++i) ptrEnum->SetIntValue( i & 0x3 );

    CPPUNIT_ASSERT_EQUAL( numCalls, CallbackUtility::Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, O1.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, O2.Count() );

}

// ---------------------------------------------------------------------------
// void CallbackTestSuite::TestCallback06()
// ---------------------------------------------------------------------------
/*!
 * Perform basic tests on a float with static callback function
 */
void CallbackTestSuite::TestCallback06()
{
    // create and initilize node map
    gcstring CameraDescription =
        CDFHeader +

        "<Float Name=\"A\"> \n"
        "  <pValue>B</pValue> \n"
        "  <pMin>C</pMin> \n"
        "  <pMax>D</pMax> \n"
        "</Float> \n"

        "<FloatReg Name=\"B\"> \n"
        "  <pIsAvailable>E</pIsAvailable> \n"
        "  <Address>0</Address> \n"
        "  <Length>4</Length> "\
        "  <AccessMode>RW</AccessMode> \n"
        "  <pPort>MyPort</pPort> \n"
        "  <Endianess>BigEndian</Endianess> \n"
        "</FloatReg> \n"

        "<FloatReg Name=\"C\"> \n"
        "  <pIsAvailable>E</pIsAvailable> \n"
        "  <Address>4</Address> \n"
        "  <Length>4</Length> "\
        "  <AccessMode>RW</AccessMode> \n"
        "  <pPort>MyPort</pPort> \n"
        "  <Endianess>BigEndian</Endianess> \n"
        "</FloatReg> \n"

        "<FloatReg Name=\"D\"> \n"
        "  <pIsAvailable>E</pIsAvailable> \n"
        "  <Address>8</Address> \n"
        "  <Length>4</Length> "\
        "  <AccessMode>RW</AccessMode> \n"
        "  <pPort>MyPort</pPort> \n"
        "  <Endianess>BigEndian</Endianess> \n"
        "</FloatReg> \n"

        "<Integer Name=\"E\"> \n"
        "   <Value>1</Value> \n"
        "</Integer> \n"

        "<Port Name=\"MyPort\" /> \n"

        + CDFFooter;

    CNodeMapRef Camera;
    Camera._LoadXMLFromString(CameraDescription);

    // create and initialize a test port
    CTestPort Port;
    Port.CreateEntry(0x000, 42ULL, RW);
    Port.CreateEntry(0x004, 42ULL, RW);
    Port.CreateEntry(0x008, 42ULL, RW);

    // conntect the node map to the port
    Camera._Connect(&Port, "MyPort");

    CallbackUtility::Reset();

    CFloatPtr ptrA(Camera._GetNode( "A" ));
    CPPUNIT_ASSERT( ptrA.IsValid() );
    CFloatPtr ptrB(Camera._GetNode( "B" ));
    CPPUNIT_ASSERT( ptrB.IsValid() );
    CFloatPtr ptrC(Camera._GetNode( "C" ));
    CPPUNIT_ASSERT( ptrC.IsValid() );
    CFloatPtr ptrD(Camera._GetNode( "D" ));
    CPPUNIT_ASSERT( ptrD.IsValid() );

    Register( ptrA->GetNode(), &CallbackUtility::Callback );
    Register( ptrB->GetNode(), &CallbackUtility::Callback );
    Register( ptrC->GetNode(), &CallbackUtility::Callback );
    Register( ptrD->GetNode(), &CallbackUtility::Callback );

    ptrC->SetValue( 0.0 );
    ptrD->SetValue( 1.0 );
    uint32_t numCalls;

    CallbackUtility::Reset();
    ptrA->SetValue( 0.5 );
    numCalls = 2;
    CPPUNIT_ASSERT_EQUAL( numCalls, CallbackUtility::Count() );

    CallbackUtility::Reset();
    ptrB->SetValue( 0.5 );
    numCalls = 2;
    CPPUNIT_ASSERT_EQUAL( numCalls, CallbackUtility::Count() );

    CallbackUtility::Reset();
    ptrC->SetValue( -1.0 );
    numCalls = 2;
    CPPUNIT_ASSERT_EQUAL( numCalls, CallbackUtility::Count() );

    CallbackUtility::Reset();
    ptrD->SetValue( 2.5 );
    numCalls = 2;
    CPPUNIT_ASSERT_EQUAL( numCalls, CallbackUtility::Count() );

    CallbackUtility::Reset();
    ptrD->GetValue();
    numCalls = 0;
    CPPUNIT_ASSERT_EQUAL( numCalls, CallbackUtility::Count() );

    numCalls = 2;
    ptrD->GetNode()->InvalidateNode();
    CPPUNIT_ASSERT_EQUAL( numCalls, CallbackUtility::Count() );

    CallbackUtility::Reset();
    CIntegerPtr ptrE = Camera._GetNode( "E" );
    numCalls = 4;
    ptrE->SetValue( 0 );
    CPPUNIT_ASSERT_EQUAL( numCalls, CallbackUtility::Count() );

}

// ---------------------------------------------------------------------------
// void CallbackTestSuite::TestCallback07()
// ---------------------------------------------------------------------------
/*!
 * Perform basic tests on Enums with two callback objects and the callbackutility
 */
void CallbackTestSuite::TestCallback07()
{
    // create and initilize node map
    gcstring CameraDescription =
        CDFHeader +

        "<Enumeration Name=\"MyFood\"> \n"
        "  <EnumEntry Name=\"Orange\" > \n"
        "    <Value>0</Value> \n"
        "  </EnumEntry> \n"
        "  <EnumEntry Name=\"Apple\" > \n"
        "    <Value>1</Value> \n"
        "  </EnumEntry> \n"
        "  <EnumEntry Name=\"Pretzel\" > \n"
        "    <Value>2</Value> \n"
        "  </EnumEntry> \n"
        "  <EnumEntry Name=\"Juice\" > \n"
        "    <Value>3</Value> \n"
        "  </EnumEntry> \n"
        "  <pValue>curFood</pValue> \n"
        "  <pSelected>curDishPrice</pSelected> \n"
        "  <pSelected>GrandTotal</pSelected> \n"
        "</Enumeration> \n"

        "<Integer Name=\"Price\"> \n"
        "  <pValue>GrandTotal</pValue>\n"
        "</Integer> \n"

        "<Integer Name=\"NumPersons\"> \n"
        "  <Value>5</Value> \n"
        "</Integer> \n"

        "<IntReg Name=\"curFood\"> \n"
        "  <Address>0</Address> \n"
        "  <Length>4</Length> \n"
        "  <AccessMode>RW</AccessMode> \n"
        "  <pPort>MyPort</pPort> \n"
        "  <Sign>Unsigned</Sign> \n"
        "  <Endianess>BigEndian</Endianess> \n"
        "</IntReg> \n"

        "<IntReg Name=\"curDishPrice\"> \n"
        "  <Address>8</Address> \n"
        "  <pAddress>curFood</pAddress> \n"
        "  <Length>4</Length> \n"
        "  <AccessMode>RW</AccessMode> \n"
        "  <pPort>MyPort</pPort> \n"
        "  <Sign>Unsigned</Sign> \n"
        "  <Endianess>BigEndian</Endianess> \n"
        "</IntReg> \n"

        "<IntSwissKnife Name=\"GrandTotal\"> \n"
        "  <pVariable Name=\"N\">NumPersons</pVariable> \n"
        "  <pVariable Name=\"P\">curDishPrice</pVariable> \n"
        "  <Formula>N*P</Formula> \n"
        "</IntSwissKnife> \n"

        "<Port Name=\"MyPort\" /> \n"

        + CDFFooter;

    CNodeMapRef Camera;
    Camera._LoadXMLFromString(CameraDescription);

    // create and initialize a test port
    CTestPort Port;
    Port.CreateEntry(0x000, 42ULL, RW);
    Port.CreateEntry(0x004, 10UL, RW);
    Port.CreateEntry(0x008, 20UL, RW);
    Port.CreateEntry(0x009, 40UL, RW);
    Port.CreateEntry(0x00a, 50UL, RW);
    Port.CreateEntry(0x00b, 60UL, RW);

    // conntect the node map to the port
    Camera._Connect(&Port, "MyPort");
    CallbackObject cbPrice, cbFood, cbPerson;

    CEnumerationPtr ptrEnum = Camera._GetNode( "MyFood" );
    CPPUNIT_ASSERT( ptrEnum.IsValid() );

    CIntegerPtr ptrFood = Camera._GetNode( "curFood" );
    CPPUNIT_ASSERT( ptrFood.IsValid() );

    CIntegerPtr ptrPrice = Camera._GetNode( "Price" );
    CPPUNIT_ASSERT( ptrPrice.IsValid() );

    CIntegerPtr ptrDishPrice = Camera._GetNode( "curDishPrice" );
    CPPUNIT_ASSERT( ptrDishPrice.IsValid() );

    CIntegerPtr ptrNumPersons = Camera._GetNode( "NumPersons" );
    CPPUNIT_ASSERT( ptrNumPersons.IsValid() );

    Register( ptrEnum->GetNode(), cbFood, &CallbackObject::Callback );
    Register( ptrPrice->GetNode(), cbPrice, &CallbackObject::Callback );
    Register( ptrNumPersons->GetNode(), cbPerson, &CallbackObject::Callback );

    uint32_t numCalls = 1;


    // change the number of persons -> Price changes, food not
    cbFood.Reset();
    cbPrice.Reset();
    cbPerson.Reset();
    ptrNumPersons->SetValue( 2 );
    numCalls = 0;
    CPPUNIT_ASSERT_EQUAL( numCalls, cbFood.Count() );
    numCalls = 1;
    CPPUNIT_ASSERT_EQUAL( numCalls, cbPrice.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbPerson.Count() );

    // change the food -> Price changes, number of persons not
    cbFood.Reset();
    cbPrice.Reset();
    cbPerson.Reset();
    ptrFood->SetValue( 2 );
    numCalls = 0;
    CPPUNIT_ASSERT_EQUAL( numCalls, cbPerson.Count() );
    numCalls = 1;
    CPPUNIT_ASSERT_EQUAL( numCalls, cbFood.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbPrice.Count() );

    // change the dish price -> Price changes, number of persons and food not
    cbFood.Reset();
    cbPrice.Reset();
    cbPerson.Reset();
    ptrDishPrice->SetValue( 2 );
    numCalls = 0;
    CPPUNIT_ASSERT_EQUAL( numCalls, cbPerson.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbFood.Count() );
    numCalls = 1;
    CPPUNIT_ASSERT_EQUAL( numCalls, cbPrice.Count() );



}

// ---------------------------------------------------------------------------
// void CallbackTestSuite::TestCallback08()
// ---------------------------------------------------------------------------
/*!
 * Perform  tests selector and invalidator
 */
void CallbackTestSuite::TestCallback08()
{
    // create and initilize node map
    gcstring CameraDescription =
        CDFHeader +

        "<Integer Name=\"A\"> \n"
        "  <pValue>D</pValue>\n"
        "</Integer> \n"

        "<Integer Name=\"B\"> \n"
        "  <pValue>F</pValue> \n"
        "</Integer> \n"

        "<Float Name=\"C\"> \n"
        "  <Value>3.141</Value>\n"
        "</Float>"

        "<MaskedIntReg Name=\"D\" >\n"
        "  <Address>0</Address> \n"
        "  <Length>4</Length> \n"
        "  <AccessMode>RW</AccessMode> \n"
        "  <pPort>MyPort</pPort> \n"
        "  <pInvalidator>E</pInvalidator> \n"
        "  <LSB>15</LSB> \n"
        "  <MSB>8</MSB> \n"
        "  <Sign>Unsigned</Sign> \n"
        "  <Endianess>BigEndian</Endianess> \n"
        "  <pSelected>F</pSelected> \n"
        "</MaskedIntReg>\n"

        "<IntReg Name=\"E\"> \n"
        "  <Address>0</Address> \n"
        "  <Length>4</Length> \n"
        "  <AccessMode>RW</AccessMode> \n"
        "  <pPort>MyPort</pPort> \n"
        "  <pInvalidator>D</pInvalidator> \n"
        "  <Sign>Unsigned</Sign> \n"
        "  <Endianess>BigEndian</Endianess> \n"
        "</IntReg> \n"

        "<IntReg Name=\"F\"> \n"
        "  <Address>4</Address> \n"
        "  <Length>4</Length> \n"
        "  <AccessMode>RW</AccessMode> \n"
        "  <pPort>MyPort</pPort> \n"
        "  <Sign>Unsigned</Sign> \n"
        "  <Endianess>BigEndian</Endianess> \n"
        "</IntReg> \n"

        "<IntReg Name=\"G\"> \n"
        "  <Address>8</Address> \n"
        "  <Length>4</Length> \n"
        "  <AccessMode>RW</AccessMode> \n"
        "  <pPort>MyPort</pPort> \n"
        "  <pInvalidator>E</pInvalidator> \n"
        "  <Sign>Unsigned</Sign> \n"
        "  <Endianess>BigEndian</Endianess> \n"
        "</IntReg> \n"

        "<Port Name=\"MyPort\" /> \n"

        + CDFFooter;

    CNodeMapRef Camera;
    Camera._LoadXMLFromString(CameraDescription);

    // create and initialize a test port
    CTestPort Port;
    Port.CreateEntry(0x000, 42ULL, RW);
    Port.CreateEntry(0x004, 10UL, RW);
    Port.CreateEntry(0x008, 20UL, RW);
    Port.CreateEntry(0x009, 40UL, RW);
    Port.CreateEntry(0x00a, 50UL, RW);
    Port.CreateEntry(0x00b, 60UL, RW);

    // conntect the node map to the port
    Camera._Connect(&Port, "MyPort");

    CallbackObject cbA, cbB, cbC, cbD, cbE, cbF, cbG;

    CIntegerPtr ptrA = Camera._GetNode( "A" );
    CIntegerPtr ptrB = Camera._GetNode( "B" );
    CFloatPtr ptrC = Camera._GetNode( "C" );
    CIntegerPtr ptrD = Camera._GetNode( "D" );
    CIntegerPtr ptrE = Camera._GetNode( "E" );
    CIntegerPtr ptrF = Camera._GetNode( "F" );
    CIntegerPtr ptrG = Camera._GetNode( "G" );
    CPPUNIT_ASSERT( ptrA.IsValid() );
    CPPUNIT_ASSERT( ptrB.IsValid() );
    CPPUNIT_ASSERT( ptrC.IsValid() );
    CPPUNIT_ASSERT( ptrD.IsValid() );
    CPPUNIT_ASSERT( ptrE.IsValid() );
    CPPUNIT_ASSERT( ptrF.IsValid() );
    CPPUNIT_ASSERT( ptrG.IsValid() );

    Register( ptrA->GetNode(), cbA, &CallbackObject::Callback );
    Register( ptrB->GetNode(), cbB, &CallbackObject::Callback );
    Register( ptrC->GetNode(), cbC, &CallbackObject::Callback );
    Register( ptrD->GetNode(), cbD, &CallbackObject::Callback );
    Register( ptrE->GetNode(), cbE, &CallbackObject::Callback );
    Register( ptrF->GetNode(), cbF, &CallbackObject::Callback );
    Register( ptrG->GetNode(), cbG, &CallbackObject::Callback );

    // using a variable to help compiler deduce the correct function
    uint32_t numCalls;


    // set A => cb of A, B, D, E; F, G; not C
    numCalls = 1;
    cbA.Reset();
    cbA.Reset();
    cbB.Reset();
    cbC.Reset();
    cbD.Reset();
    cbE.Reset();
    cbF.Reset();
    cbG.Reset();
    ptrA->SetValue( 0 );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbA.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbB.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbD.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbE.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbF.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbG.Count() );
    numCalls = 0;
    CPPUNIT_ASSERT_EQUAL( numCalls, cbC.Count() );

    // set B => cb of B, F; not A, C, D, E, G
    numCalls = 1;
    cbA.Reset();
    cbA.Reset();
    cbB.Reset();
    cbC.Reset();
    cbD.Reset();
    cbE.Reset();
    cbF.Reset();
    cbG.Reset();
    ptrB->SetValue( 0 );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbB.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbB.Count() );
    numCalls = 0;
    CPPUNIT_ASSERT_EQUAL( numCalls, cbA.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbC.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbD.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbE.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbG.Count() );

    // set C => no cb
    numCalls = 1;
    cbA.Reset();
    cbA.Reset();
    cbB.Reset();
    cbC.Reset();
    cbD.Reset();
    cbE.Reset();
    cbF.Reset();
    cbG.Reset();
    ptrC->SetValue( 0 );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbC.Count() );
    numCalls = 0;
    CPPUNIT_ASSERT_EQUAL( numCalls, cbA.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbB.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbD.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbE.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbF.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbG.Count() );

    // set D => cb of A, B, D, F, E, G, C
    numCalls = 1;
    cbA.Reset();
    cbA.Reset();
    cbB.Reset();
    cbC.Reset();
    cbD.Reset();
    cbE.Reset();
    cbF.Reset();
    cbG.Reset();
    ptrD->SetValue( 0 );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbA.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbB.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbD.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbE.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbF.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbG.Count() );
    numCalls = 0;
    CPPUNIT_ASSERT_EQUAL( numCalls, cbC.Count() );

    // set E => cb of A, B, D, F, E, G, C
    numCalls = 1;
    cbA.Reset();
    cbA.Reset();
    cbB.Reset();
    cbC.Reset();
    cbD.Reset();
    cbE.Reset();
    cbF.Reset();
    cbG.Reset();
    ptrE->SetValue( 0 );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbA.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbB.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbD.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbE.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbF.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbG.Count() );
    numCalls = 0;
    CPPUNIT_ASSERT_EQUAL( numCalls, cbC.Count() );

    // set F => cb of B, F; not of A, B, C, D, E,
    numCalls = 1;
    cbA.Reset();
    cbA.Reset();
    cbB.Reset();
    cbC.Reset();
    cbD.Reset();
    cbE.Reset();
    cbF.Reset();
    cbG.Reset();
    ptrF->SetValue( 0 );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbB.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbF.Count() );
    numCalls = 0;
    CPPUNIT_ASSERT_EQUAL( numCalls, cbA.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbC.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbD.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbE.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbG.Count() );

    // set G => cb of G; not of A, B, C, D, E, F
    numCalls = 1;
    cbA.Reset();
    cbA.Reset();
    cbB.Reset();
    cbC.Reset();
    cbD.Reset();
    cbE.Reset();
    cbF.Reset();
    cbG.Reset();
    ptrG->SetValue( 0 );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbG.Count() );
    numCalls = 0;
    CPPUNIT_ASSERT_EQUAL( numCalls, cbA.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbB.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbC.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbD.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbE.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbF.Count() );
}


// ---------------------------------------------------------------------------
// void CallbackTestSuite::TestCallback09()
// ---------------------------------------------------------------------------
/*!
 * \brief Perform tests on DCAM-like feature block
 *
 * The feature contains two numerical values ("C" and "D") and a read-only flag
 * ("A") signalling the presence. The base address is obtained from a SmartFeature
 * register ("C"). This register is enabled by unlocking the Advanced Feature
 * Register ("F").
 *
 *
 * Several tricks are used to mimic the camera.
 * - the smart feature address is directly set in the register space (see sfvalue
 *   bitset) and matches the value in XML description
 * - the port object handles the hardcoded address (0) differently and returns
 *   0 or -1 depending on an internal state.
 */
void CallbackTestSuite::TestCallback09()
{
    class CMyTestPort : public CTestPort
    {
        bool m_Enable;
    public:
        CMyTestPort() : m_Enable(false){}
        void EnableAdvFeature( bool enable=true ) {m_Enable=enable;}
        virtual void Write(const void *pBuffer, int64_t Address, int64_t Length)
        {
            if (Address>=0 && Address<8)
                0;
            else
                CTestPort::Write( pBuffer, Address, Length );
        }
        virtual void Read( void *pBuffer, int64_t Address, int64_t Length )
        {
            if (Address>=0 && Address<8)
                memset( pBuffer, m_Enable ? 0 : -1 , (size_t)Length );
            else
                CTestPort::Read( pBuffer, Address, Length );
        }
    };

    // create and initilize node map
    gcstring CameraDescription =
        CDFHeader +

        "<StructReg Comment=\"Fancy Feature\" > \n"

        "  <pIsAvailable>B</pIsAvailable> \n"
        "  <pAddress>B</pAddress> \n"
        "  <Length>4</Length> \n"
        "  <AccessMode>RW</AccessMode> \n"
        "  <pPort>MyPort</pPort> \n"
        "  <Endianess>LittleEndian</Endianess>    \n"

        // bool meaning Feature Available
        "  <StructEntry Name=\"A\"> \n"
        "    <AccessMode>RO</AccessMode> \n"
        "    <Bit>0</Bit> \n"
        "  </StructEntry> \n"

        // Value 1
        "  <StructEntry Name=\"C\" > \n"
        "    <pIsAvailable>A</pIsAvailable> \n"
        "    <LSB>8</LSB> \n"
        "    <MSB>15</MSB> \n"
        "  </StructEntry> \n"

        // Value 2
        "  <StructEntry Name=\"D\" > \n"
        "    <pIsAvailable>A</pIsAvailable> \n"
        "    <LSB>16</LSB> \n"
        "    <MSB>23</MSB> \n"
        "  </StructEntry> \n"

        "</StructReg> \n"

        // Base-Address of Fancy
        "<SmartFeature Name=\"B\"> \n"
        "  <FeatureID>A014EDBC-6768-4bcb-B039-9070961F2843</FeatureID> \n"
        "  <Address>8</Address> \n"
        "  <pPort>MyPort</pPort> \n"
        "  <pIsAvailable>F</pIsAvailable> \n"
        "</SmartFeature> \n"

        // DCAM-like lock
        "<AdvFeatureLock Name=\"F\"> \n"
        "  <Address>0</Address> \n"
        "  <Length>8</Length> \n"
        "  <AccessMode>RW</AccessMode> \n"
        "  <pPort>MyPort</pPort> \n"
        "  <FeatureID>0x0030533B73C3</FeatureID> \n"
        "  <Timeout>0</Timeout> \n"
        "</AdvFeatureLock> \n"

        "<Port Name=\"MyPort\" /> \n"

        + CDFFooter;

    CNodeMapRef Camera;
    Camera._LoadXMLFromString(CameraDescription);

    // create and initialize a test port
    CMyTestPort Port;
    std::bitset<24*8> sfvalue;
     sfvalue.set(); // all bits set => feature is not available
    Port.CreateEntry(0x000, (unsigned long)-1L, RW);
    Port.CreateEntry(0x004, (unsigned long)-1L, RW);
    Port.CreateEntry(0x008, 24, &sfvalue, RW);
    Port.CreateEntry(0x020, (unsigned long)-1L, RW);

    // conntect the node map to the port
    Camera._Connect(&Port, "MyPort");

    CallbackObject cbA, cbB, cbC, cbD, cbE, cbF, cbG;

    CIntegerPtr ptrA = Camera._GetNode( "A" );
    CIntegerPtr ptrB = Camera._GetNode( "B" );
    CIntegerPtr ptrC = Camera._GetNode( "C" );
    CIntegerPtr ptrD = Camera._GetNode( "D" );
    CIntegerPtr ptrF = Camera._GetNode( "F" );
    CPPUNIT_ASSERT( ptrA.IsValid() );
    CPPUNIT_ASSERT( ptrB.IsValid() );
    CPPUNIT_ASSERT( ptrC.IsValid() );
    CPPUNIT_ASSERT( ptrD.IsValid() );
    CPPUNIT_ASSERT( ptrF.IsValid() );

    Register( ptrA->GetNode(), cbA, &CallbackObject::Callback );
    Register( ptrB->GetNode(), cbB, &CallbackObject::Callback );
    Register( ptrC->GetNode(), cbC, &CallbackObject::Callback );
    Register( ptrD->GetNode(), cbD, &CallbackObject::Callback );
    Register( ptrF->GetNode(), cbF, &CallbackObject::Callback );

    // using a variable to help compiler deduce the correct function
    uint32_t numCalls;


    CPPUNIT_ASSERT_EQUAL( RO, ptrF->GetAccessMode() );
    CPPUNIT_ASSERT_EQUAL( NA, ptrB->GetAccessMode() );
    CPPUNIT_ASSERT_EQUAL( NA, ptrA->GetAccessMode() );
    CPPUNIT_ASSERT_EQUAL( NA, ptrC->GetAccessMode() );
    CPPUNIT_ASSERT_EQUAL( NA, ptrD->GetAccessMode() );

    // enable advanced feature and set address 0x20in quad 5
    Port.EnableAdvFeature();
    sfvalue.reset();
    sfvalue.set( 19*8+5 );
    Port.Write( &sfvalue, 0x8, 24 );
    ptrF->GetNode()->InvalidateNode(); // invalidate advanced feature lock F
    ptrB->GetNode()->InvalidateNode(); // invalidate smart feature B
    cbA.Reset(), cbB.Reset(), cbC.Reset(), cbD.Reset(), cbE.Reset(), cbF.Reset();

    CPPUNIT_ASSERT_EQUAL( RO, ptrB->GetAccessMode() );
    CPPUNIT_ASSERT_EQUAL( RO, ptrA->GetAccessMode() );
    CPPUNIT_ASSERT_EQUAL( RW, ptrC->GetAccessMode() );
    CPPUNIT_ASSERT_EQUAL( RW, ptrD->GetAccessMode() );

    // no callbacks although B and F write implicitly
    numCalls = 0;
    CPPUNIT_ASSERT_EQUAL( numCalls, cbA.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbB.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbC.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbD.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbF.Count() );

    // Set feature C => no callbacks but C itself
    *ptrC = 0;
    numCalls = 0;
    CPPUNIT_ASSERT_EQUAL( numCalls, cbA.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbB.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbD.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbF.Count() );
    numCalls = 1;
    CPPUNIT_ASSERT_EQUAL( numCalls, cbC.Count() );
    cbC.Reset();

    // Set feature D => no callbacks but D itself
    *ptrD = 0;
    numCalls = 0;
    CPPUNIT_ASSERT_EQUAL( numCalls, cbA.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbB.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbC.Count() );
    CPPUNIT_ASSERT_EQUAL( numCalls, cbF.Count() );
    numCalls = 1;
    CPPUNIT_ASSERT_EQUAL( numCalls, cbD.Count() );

}
// ---------------------------------------------------------------------------
// void CallbackTestSuite::TestCounter
// ---------------------------------------------------------------------------
/*!
 * Tests the counter class
 */

void CallbackTestSuite::TestCounter()
{
    GenApi::Counter c;

    CPPUNIT_ASSERT_EQUAL( (unsigned int) 0, (unsigned int) c);
    CPPUNIT_ASSERT(c.IsZero());

    CPPUNIT_ASSERT_EQUAL((unsigned int) 0, c++);
    CPPUNIT_ASSERT(!c.IsZero());
    CPPUNIT_ASSERT_EQUAL((unsigned int) 1, (unsigned int) c);

    CPPUNIT_ASSERT_EQUAL((unsigned int) 2, (unsigned int ) ++c);
    CPPUNIT_ASSERT_EQUAL((unsigned int) 2, (unsigned int ) c);


    CPPUNIT_ASSERT_EQUAL((unsigned int) 2, (unsigned int ) c--);
    CPPUNIT_ASSERT_EQUAL((unsigned int) 1, (unsigned int) c);
    CPPUNIT_ASSERT_EQUAL((unsigned int) 0, (unsigned int ) --c);

    CPPUNIT_ASSERT_EQUAL((unsigned int) 0, (unsigned int) c );

    CPPUNIT_ASSERT_EQUAL((unsigned int) 0, c.GetValue());
}





// ---------------------------------------------------------------------------
// void CallbackTestSuite::TestCallback10()
// ---------------------------------------------------------------------------
/*!
 * Checks callbacks triggered by InvalidateNode() & SetValue().
 */

void CallbackTestSuite::TestCallback10()
{
    // create and initilize node map
    gcstring CameraDescription =
        CDFHeader +

        "<Float Name=\"A\"> \n"
        "  <pValue>B</pValue> \n"
        "  <pMin>C</pMin> \n"
        "  <pMax>D</pMax> \n"
        "</Float> \n"

        "<FloatReg Name=\"B\"> \n"
        "  <pIsAvailable>E</pIsAvailable> \n"
        "  <Address>0</Address> \n"
        "  <Length>4</Length> "\
        "  <AccessMode>RW</AccessMode> \n"
        "  <pPort>MyPort</pPort> \n"
        "  <Endianess>BigEndian</Endianess> \n"
        "</FloatReg> \n"

        "<FloatReg Name=\"C\"> \n"
        "  <pIsAvailable>E</pIsAvailable> \n"
        "  <Address>4</Address> \n"
        "  <Length>4</Length> "\
        "  <AccessMode>RW</AccessMode> \n"
        "  <pPort>MyPort</pPort> \n"
        "  <Endianess>BigEndian</Endianess> \n"
        "</FloatReg> \n"

        "<FloatReg Name=\"D\"> \n"
        "  <pIsAvailable>E</pIsAvailable> \n"
        "  <Address>8</Address> \n"
        "  <Length>4</Length> "\
        "  <AccessMode>RW</AccessMode> \n"
        "  <pPort>MyPort</pPort> \n"
        "  <Endianess>BigEndian</Endianess> \n"
        "</FloatReg> \n"

        "<Integer Name=\"E\"> \n"
        "   <Value>1</Value> \n"
        "</Integer> \n"

        "<Port Name=\"MyPort\" /> \n"

        + CDFFooter;

    CNodeMapRef Camera;
    Camera._LoadXMLFromString(CameraDescription);

    // create and initialize a test port
    CTestPort Port;
    Port.CreateEntry(0x000, 42ULL, RW);
    Port.CreateEntry(0x004, 42ULL, RW);
    Port.CreateEntry(0x008, 42ULL, RW);

    // conntect the node map to the port
    Camera._Connect(&Port, "MyPort");

    CFloatPtr ptrA = Camera._GetNode( "A" );
    CPPUNIT_ASSERT( ptrA.IsValid() );
    CFloatPtr ptrB = Camera._GetNode( "B" );
    CPPUNIT_ASSERT( ptrB.IsValid() );
    CFloatPtr ptrC = Camera._GetNode( "C" );
    CPPUNIT_ASSERT( ptrC.IsValid() );
    CFloatPtr ptrD = Camera._GetNode( "D" );
    CPPUNIT_ASSERT( ptrD.IsValid() );
    CIntegerPtr ptrE = Camera._GetNode( "E" );
    CPPUNIT_ASSERT( ptrE.IsValid() );
    CNodePtr ptrMyPort = Camera._GetNode( "MyPort" );
    CPPUNIT_ASSERT( ptrMyPort.IsValid() );

    CallbackObject obA, obB, obC, obD, obE, obMyPort;
    Register( ptrA->GetNode(), obA, &CallbackObject::Callback );
    Register( ptrB->GetNode(), obB, &CallbackObject::Callback );
    Register( ptrC->GetNode(), obC, &CallbackObject::Callback );
    Register( ptrD->GetNode(), obD, &CallbackObject::Callback );
    Register( ptrE->GetNode(), obE, &CallbackObject::Callback );
    Register( ptrMyPort, obMyPort, &CallbackObject::Callback );

    // Check number of callbacks triggered by SetValue
    // Use C und D first, because they serve as pMin and pMax for A
    obA.Reset(); obB.Reset(); obC.Reset(); obD.Reset(); obE.Reset(); obMyPort.Reset();
    ptrC->SetValue( 0.0 );
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obA.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obB.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obC.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obD.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obE.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obMyPort.Count());

    obA.Reset(); obB.Reset(); obC.Reset(); obD.Reset(); obE.Reset(); obMyPort.Reset();
    ptrD->SetValue( 1.0 );
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obA.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obB.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obC.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obD.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obE.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obMyPort.Count());

    obA.Reset(); obB.Reset(); obC.Reset(); obD.Reset(); obE.Reset(); obMyPort.Reset();
    ptrB->SetValue( 1.0 );
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obA.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obB.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obC.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obD.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obE.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obMyPort.Count());

    obA.Reset(); obB.Reset(); obC.Reset(); obD.Reset(); obE.Reset(); obMyPort.Reset();
    ptrA->SetValue(1.0);
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obA.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obB.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obC.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obD.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obE.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obMyPort.Count());

    obA.Reset(); obB.Reset(); obC.Reset(); obD.Reset(); obE.Reset(); obMyPort.Reset();
    ptrE->SetValue(1);
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obA.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obB.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obC.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obD.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obE.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obMyPort.Count());

    obA.Reset(); obB.Reset(); obC.Reset(); obD.Reset(); obE.Reset(); obMyPort.Reset();
    ptrA->GetNode()->InvalidateNode();
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obA.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obB.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obC.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obD.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obE.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obMyPort.Count());

    obA.Reset(); obB.Reset(); obC.Reset(); obD.Reset(); obE.Reset(); obMyPort.Reset();
    ptrB->GetNode()->InvalidateNode();
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obA.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obB.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obC.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obD.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obE.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obMyPort.Count());

    obA.Reset(); obB.Reset(); obC.Reset(); obD.Reset(); obE.Reset(); obMyPort.Reset();
    ptrC->GetNode()->InvalidateNode();
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obA.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obB.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obC.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obD.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obE.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obMyPort.Count());

    obA.Reset(); obB.Reset(); obC.Reset(); obD.Reset(); obE.Reset(); obMyPort.Reset();
    ptrD->GetNode()->InvalidateNode();
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obA.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obB.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obC.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obD.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obE.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obMyPort.Count());

    obA.Reset(); obB.Reset(); obC.Reset(); obD.Reset(); obE.Reset(); obMyPort.Reset();
    ptrE->GetNode()->InvalidateNode();
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obA.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obB.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obC.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obD.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obE.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obMyPort.Count());

    obA.Reset(); obB.Reset(); obC.Reset(); obD.Reset(); obE.Reset(); obMyPort.Reset();
    ptrMyPort->InvalidateNode();
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obA.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obB.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obC.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obD.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 0, obE.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obMyPort.Count());

    obA.Reset(); obB.Reset(); obC.Reset(); obD.Reset(); obE.Reset(); obMyPort.Reset();
    Camera._InvalidateNodes();
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obA.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obB.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obC.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obD.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obE.Count());
    CPPUNIT_ASSERT_EQUAL( (uint32_t) 1, obMyPort.Count());




}

// ---------------------------------------------------------------------------
// void CallbackTestSuite::TestCallback11()
// ---------------------------------------------------------------------------
/*!
 * Perform basic tests on a float with static callback function
 */
void CallbackTestSuite::TestCallback11()
{
    // create and initilize node map
    gcstring CameraDescription =
        CDFHeader +

        "<StringReg Name=\"MyString\"> \n"
        "  <Address>4</Address> \n"
        "  <pLength>A</pLength> \n"
        "  <AccessMode>RW</AccessMode> \n"
        "  <pPort>MyPort</pPort> \n"
        "</StringReg> \n"

        "<IntReg Name=\"A\" >\n"
        "  <Address>0</Address> \n"
        "  <Length>4</Length> \n"
        "  <AccessMode>RW</AccessMode> \n"
        "  <pPort>MyPort</pPort> \n"
        "  <Sign>Unsigned</Sign> \n"
        "  <Endianess>LittleEndian</Endianess> \n"
        "</IntReg> \n"

        "<Port Name=\"MyPort\" /> \n"

        + CDFFooter;

    CNodeMapRef Camera;
    Camera._LoadXMLFromString(CameraDescription);

    // create and initialize a test port
    CTestPort Port;
    const char testvalue[8] = "foobar";
    Port.CreateEntry(0x000, 7UL, RW);
    Port.CreateEntry(0x004, 7, (void*)(testvalue+0), RW);
    Port.CreateEntry(0x008, 4, (void*)(testvalue+4), RW);

    // conntect the node map to the port
    Camera._Connect(&Port, "MyPort");

    CallbackUtility::Reset();

    CStringPtr ptrString = Camera._GetNode( "MyString" );
    CIntegerPtr ptrLength = Camera._GetNode( "A" );
    CPPUNIT_ASSERT( ptrString.IsValid() );
    CPPUNIT_ASSERT( ptrLength.IsValid() );

    Register( ptrString->GetNode(), &CallbackUtility::Callback );

    CPPUNIT_ASSERT_EQUAL( gcstring( testvalue ), ptrString->GetValue() );


    uint32_t numCalls;

    // change the length => callback of string
    *ptrLength = 4;
    numCalls = 1;
    CPPUNIT_ASSERT_EQUAL( numCalls, CallbackUtility::Count() );
    gcstring g( testvalue );
    CPPUNIT_ASSERT_EQUAL( g.substr( 0, 4 ), ptrString->GetValue() );
    CRegisterPtr ptrStringReg(ptrString);
    CPPUNIT_ASSERT_EQUAL( (int64_t)4, ptrStringReg->GetLength() );

}

void CallbackTestSuite::TestCallback12()
{
    // create and initialize node map
    gcstring CameraDescription =
        CDFHeader +

        "<Category Name=\"Event\">      \n"
        "  <pFeature>Node</pFeature>    \n"
        "</Category>                    \n"

        "<Integer Name=\"Node\">        \n"
        "  <Value>1</Value>             \n"
        "</Integer>                     \n"

        + CDFFooter;

    CNodeMapRef Camera;
    Camera._LoadXMLFromString(CameraDescription);

    CIntegerPtr ptrNode = Camera._GetNode( "Node" );
    CPPUNIT_ASSERT( ptrNode.IsValid() );

    CallbackObject cbEvent;
    Register( ptrNode->GetNode(), cbEvent, &CallbackObject::Callback );

    CPPUNIT_ASSERT_EQUAL( (uint32_t)0, cbEvent.Count() );
    ptrNode->SetValue( 0 );
    CPPUNIT_ASSERT_EQUAL( (uint32_t)1, cbEvent.Count() );

}


void Callback13_inside( INode *pNode )
{
#if defined (WIN32)

    // Never(!) use CLockEx in client code; this is for test only
    CLockEx &Lock( *static_cast<CLockEx*>(&(pNode->GetNodeMap()->GetLock())) );

    // We must be inside the critical section
    CPPUNIT_ASSERT( Lock.GetRecursionCount() > 0 );

#endif
}

void Callback13_outside( INode *pNode )
{
#if defined (WIN32)

    // Never(!) use CLockEx in client code; this is for test only
    CLockEx &Lock( *static_cast<CLockEx*>(&(pNode->GetNodeMap()->GetLock())) );

    // We must be outside the critical section
    CPPUNIT_ASSERT( Lock.GetRecursionCount() == 0 );

#endif
}

void CallbackTestSuite::TestCallback13()
{
    // create and initialize node map
    gcstring CameraDescription =
        CDFHeader +

        "<Integer Name=\"Node\">        \n"
        "  <Value>1</Value>             \n"
        "</Integer>                     \n"

        + CDFFooter;

    CNodeMapRef Camera;
    Camera._LoadXMLFromString(CameraDescription);

    CIntegerPtr ptrNode = Camera._GetNode( "Node" );
    CPPUNIT_ASSERT( ptrNode.IsValid() );

    CallbackObject cbEvent;
    Register( ptrNode->GetNode(), Callback13_inside, cbPostInsideLock );
    Register( ptrNode->GetNode(), Callback13_outside, cbPostOutsideLock );

    ptrNode->SetValue( 0 );

}


void CallbackTestSuite::TestPortInvalidate()
{
    if(GenApiSchemaVersion < v1_1)
        return;

    // create and initilize node map
    gcstring CameraDescription =
        CDFHeader +

        "                                        \n"
        "<IntReg Name=\"Invalidator\">                \n"
        "    <Address>0x0000</Address>            \n"
        "    <Length>4</Length>                    \n"
        "    <AccessMode>RW</AccessMode>            \n"
        "    <pPort>Port</pPort>                    \n"
        "    <Sign>Signed</Sign>                    \n"
        "    <Endianess>LittleEndian</Endianess>    \n"
        "</IntReg>                                \n"
        "                                        \n"
        "<Integer Name=\"FloatingInvalidator\">                \n"
        "    <Value>0</Value>            \n"
        "</Integer>                                \n"
        "                                        \n"
        "<IntReg Name=\"Invalidated\">                \n"
        "    <Address>0x0004</Address>            \n"
        "    <Length>4</Length>                    \n"
        "    <AccessMode>RW</AccessMode>            \n"
        "    <pPort>Port</pPort>                    \n"
        "    <Sign>Signed</Sign>                    \n"
        "    <Endianess>LittleEndian</Endianess>    \n"
        "</IntReg>                                \n"

        "<Port Name=\"Port\" > \n"
        "    <pInvalidator>Invalidator</pInvalidator> \n"
        "    <pInvalidator>FloatingInvalidator</pInvalidator> \n"
        "</Port> \n"

        + CDFFooter;

    CNodeMapRef Camera;
    Camera._LoadXMLFromString(CameraDescription);

    // create and initialize a test port
    CTestPort Port;
    Port.CreateEntry(0x000, 42ULL, RW);
    Port.CreateEntry(0x004, 42ULL, RW);

    // conntect the node map to the port
    Camera._Connect(&Port, "Port");

    CallbackUtility::Reset();

    CIntegerPtr ptrInvalidator(Camera._GetNode( "Invalidator" ));
    CPPUNIT_ASSERT( ptrInvalidator.IsValid() );
    CIntegerPtr ptrFloatingInvalidator(Camera._GetNode( "FloatingInvalidator" ));
    CPPUNIT_ASSERT( ptrFloatingInvalidator.IsValid() );
    CIntegerPtr ptrInvalidated(Camera._GetNode( "Invalidated" ));
    CPPUNIT_ASSERT( ptrInvalidated.IsValid() );

    Register( ptrInvalidated->GetNode(), &CallbackUtility::Callback );

    // touch the invalidator
    ptrInvalidator->SetValue (80);
    // verify that the values depending on the port were invalidated
    // ie. cache bypassed
    uint32_t Value = 88;
    Port.Write (&Value, 0x04, sizeof(Value));
    CPPUNIT_ASSERT_EQUAL ((int64_t)88, ptrInvalidated->GetValue());
    // check if the callback was fired
    CPPUNIT_ASSERT_EQUAL( (uint32_t)1, CallbackUtility::Count() );

    // the same with floating invalidator
    ptrFloatingInvalidator->SetValue (80);
    Value = 99;
    Port.Write (&Value, 0x04, sizeof(Value));
    CPPUNIT_ASSERT_EQUAL ((int64_t)99, ptrInvalidated->GetValue());
    CPPUNIT_ASSERT_EQUAL( (uint32_t)2, CallbackUtility::Count() );
}
